#include "cset.hh"
#include "confparser.hh"

#include <sys/poll.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <signal.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h> /* for inet_ntoa */
#include <sys/wait.h>
#include <unistd.h>

#include <vector>
#include <cerrno>
#include <cstdio>
#include <set>
#include <utility>

/* Configuration */

static ConfParser config;
static const char cfgfilename[] = "ircproxy.cfg";

static class ConfigLoader
{
public:
    ConfigLoader()
    {
        FILE* fp = fopen(cfgfilename, "rt");
        if(!fp)
        {
            perror(cfgfilename);
            return;
        }
        char Buf[8192];
        setvbuf(fp, Buf, _IOFBF, sizeof Buf);
        config.Parse(fp);
        fclose(fp);
    }
} ConfigLoader;

static const ConfParser::Field& GetConf(const char *sect, const char *var)
{
    try
    {
        return config[sect][var];
    }
    catch(ConfParser::invalid_section sect_error)
    {
        typedef std::string errortype;
        static std::set<errortype> displayed;
        
        const errortype ErrorName(sect_error.GetSection());
        
        if(displayed.find(ErrorName) == displayed.end())
        {
            fprintf(stderr,
                "> %s: Missing section [%s]\n",
                cfgfilename,
                ErrorName.c_str()
                   );
            displayed.insert(ErrorName);
        }

        static const ConfParser::Field error;
        return error;
    }
    catch(ConfParser::invalid_field field_error)
    {
        typedef std::pair<std::string, std::string> errortype;
        static std::set<errortype> displayed;
        
        const errortype ErrorName(field_error.GetField().c_str(),
                                  field_error.GetSection().c_str());
        
        if(displayed.find(ErrorName) == displayed.end())
        {
            fprintf(stderr,
                "> %s: Missing \"%s\" in [%s]\n",
               cfgfilename,
               ErrorName.first.c_str(),/* fieldname */
               ErrorName.second.c_str()/* section name */
                   );

            displayed.insert(ErrorName);
        }

        static const ConfParser::Field error;
        return error;
    }
}

static bool IsUTF8Channel(const std::string& channame)
{
    /* Note: the first word of special_chain is ignored. */
    const ConfParser::ElemVec& elems = GetConf("settings", "special_chain").Fields();
    for(unsigned a=1; a<elems.size(); ++a)
        if(channame == elems[a].SField) return true;
    return false;
}
static const std::string GetServer()
{
    return GetConf("settings", "server_host");
}
static const unsigned GetServerPort()
{
    return GetConf("settings", "server_port");
}
static const unsigned GetListenPort()
{
    return GetConf("settings", "listen_port");
}

/* Util */

const std::vector<std::string> Split(const std::string& text, char separator)
{
    std::vector<std::string> words;
    std::string rest = text;
    for(;;)
    {
        unsigned spacepos = rest.find(separator);  
        if(spacepos == rest.npos)break;
        words.push_back(rest.substr(0, spacepos));   
        if(spacepos+1 >= rest.size()) { rest = ""; break; }
        rest = rest.substr(spacepos+1);
    }
    if(rest.size()) { words.push_back(rest); rest = ""; }
    return words;
}


/* From client(utf8) to server */
static const std::string Latin1Out(const std::string& s)
{
    return toLatin1(fromUTF8(s));
}
static const std::string UTF8Out(const std::string& s)
{
    return toUTF8(fromUTF8(s));
}
static const std::string OutChannel(const std::string& chan, const std::string& line)
{
    return IsUTF8Channel(chan) ? UTF8Out(line) : Latin1Out(line);
}


/* From server to client(utf8) */
static const std::string Latin1In(const std::string& s)
{
    return toUTF8(fromLatin1(s));
}
static const std::string UTF8In(const std::string& s)
{
    return toUTF8(fromUTF8(s));
}
static const std::string InChannel(const std::string& chan, const std::string& line)
{
    return IsUTF8Channel(chan) ? UTF8In(line) : Latin1In(line);
}

static bool ReadWrite(FILE* inf, FILE* outf)
{
    char Buf[8192];
    if(!fgets(Buf, sizeof Buf, inf)) return true;
    
    std::string line = Buf;
    
    std::string guess_channel = "";
    
    enum { in, out } guess_dir = out;

    std::vector<std::string> words = Split(line, ' ');
    for(unsigned a=0; a<4; ++a)
    {
        if(a >= words.size()) break;
        const std::string& word = words[a];
        if(word.empty()) continue;
        
        char ch = word[0];
        if(a == 0 && ch == ':')
        {
            guess_dir = in; // from server
        }
        if(ch == '#' || ch == '&' || ch == '+')
        {
            guess_channel = word;
            break;
        }
    }
    
/*
    fprintf(stderr, "guess(chan=%s, dir=%s)\n",
        guess_channel.c_str(),
        guess_dir==in ? "in" : "out");
*/
    std::string result =
        guess_dir == in ?  InChannel(guess_channel, line)
                        : OutChannel(guess_channel, line);
    
    fwrite(result.c_str(), 1, result.size(), outf);
    return false;
}

static FILE* ConnectIRC()
{
    struct sockaddr_in MyAddr;
    
    struct hostent *HostName = gethostbyname(GetServer().c_str());
    if(!HostName)
    {
        throw "gethostbyname";
    }
    
    memset(&MyAddr, 0, sizeof(MyAddr));

    MyAddr.sin_family = AF_INET;
    MyAddr.sin_port = htons(GetServerPort());
    memcpy(&MyAddr.sin_addr, HostName->h_addr, HostName->h_length);

    int Plug = socket(AF_INET, SOCK_STREAM, 0);
    if(Plug < 0)
    {
        throw "socket";
    }
    
    if(connect(Plug, (struct sockaddr*)&MyAddr, sizeof(MyAddr)) == -1)
    {
        close(Plug);
        throw "connect";
    }
    
    return fdopen(Plug, "r+");
}

static void Service(FILE* user)
{
    FILE* server = NULL;
    try
    {
        server = ConnectIRC();
        if(!server)
        {
            throw "Unknown error";
        }
    }
    catch(const char *s)
    {
        char Buf[512];
        
        std::string error = "Can not connect to ";
        error += GetServer();
        error += ':';
        sprintf(Buf, "%u", GetServerPort());
        error += Buf;
        
        error += ": ";
        error += s;
        error += ": ";
        if(errno != 0)
        {
            error += strerror(errno);
        }
        
        fprintf(user, "ERROR :%s\n", error.c_str());
        return;
    }
    
    // Buffering is not wanted
    setbuf(user,  NULL);
    setbuf(server, NULL);
    
    // Nonblocking is not wanted
    fcntl(fileno(user), F_SETFL,  fcntl(fileno(user), F_GETFL) & ~O_NONBLOCK);
    fcntl(fileno(server), F_SETFL, fcntl(fileno(server), F_GETFL) & ~O_NONBLOCK);
    
    for(;;)
    {
        struct pollfd polls[2] =
        {
            {fileno(user), POLLIN, 0},
            {fileno(server), POLLIN, 0}
        };
        int n = poll(polls, 2, -1);
        if(n < 0)
        {
            if(errno == EINTR) continue;
            break;
        }
        if(polls[0].revents & POLLIN)
        {
            bool eof = ReadWrite(user, server);
            if(eof) break;
        }
        if(polls[1].revents & POLLIN)
        {
            bool eof = ReadWrite(server, user);
            if(eof) break;
        }
    }
    
    fclose(user);
    fclose(server);
}

static void CatchDead(int)
{
    /* Wait for the dead. */
    waitpid(-1, NULL, WNOHANG);
}

int main(void)
{
    /* Close everything but stderr */
    for(int fd=0;fd<512;++fd) { if(fd != STDERR_FILENO) close(fd); }
    
    struct sockaddr_in MyAddr, They;
    
    memset(&MyAddr, 0, sizeof(MyAddr));

    MyAddr.sin_family = AF_INET;
    MyAddr.sin_port = htons(GetListenPort());
    MyAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* auto-fill with my IP, network long  */

    int Plug = socket(AF_INET, SOCK_STREAM, 0);
    if(Plug < 0) { perror("socket"); return -errno; }
    
    int flag=1; /* Just a temp variable, boolean value for setsockopt() */
    setsockopt(Plug, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof flag);
    
    if(bind(Plug, (struct sockaddr *)&MyAddr, sizeof(struct sockaddr)))
    {
        perror("bind");
        close(Plug);
        return -errno;
    }
    
    if(listen(Plug, 1))
    {
        close(Plug);
        perror("listen");
        return -errno;
    }
    
    signal(SIGCHLD, CatchDead);
    
    for(;;)
    {
        unsigned AddrLen = sizeof(They);
        int Sock = accept(Plug, (struct sockaddr*)&They, &AddrLen);
        if(Sock < 0)
        {
            if(errno == EINTR || errno == EAGAIN) continue;
            
            perror("accept");
            break;
        }
        int forkpid = fork();
        switch(forkpid)
        {
            case -1:
                perror("fork");
                break;
            case 0:
                close(Plug);
                close(STDERR_FILENO);
                signal(SIGCHLD, SIG_IGN);
                Service(fdopen(Sock, "r+"));
                close(Sock);
                _exit(0);
                break;
        }
        close(Sock);
    }
    close(Plug);
    return 0;
}
